/*

	Filename:   MemoText.cc
   Version:    1.20
   Date:       2001 01 09

   For:        VdBase 7.x / dB2K / Win9x / Win2000

   Author:     Ronnie MacGregor
   Contact:    ronnie@high-lugtonridge.co.uk
	
         
   Description:
   ************
      A custom editor control for memo text which avoids using a blob file
      (.dbt) for storage, by storing the memo text in table rows.


   Revisions:
   **********
      2001 06 09  Creation of MemoText table added to mtSetup().
                        

   Methods:
   ********
      mtSetup(cTableName, oLinkedRowset, cLinkFieldName)

         Should be called from the form's onOpen to set up the control.

         cTableName

            A string containing the name of the table to store the memo text in.
            This name should be the same as the name of the table containing the
            link field, but with a .mtf extension to signify that it is a Memo
            Text file. 

         oLinkedRowset

            An object reference to the rowset you are using for the table 
            containing the link field for the MemoText.

         cLinkFieldName

            A string containg the name of the field in the above rowset, in which
            the MemoText control will store the link reference to the memo text.
            This field should be a Character field of width 10.


      mtOnClose()

         The editor object doesn't have a native onClose event, so this event
         should be called by the form's onClose event, to enable the control to
         "try" a pack of the MemoText Table.


      The following methods of the control must be called by the corresponding
      rowset events for the table containing the memo text link field, and should
      be pretty much self explanatory.

      mtAbandon()
      mtAppend()
      mtDelete()
      mtEdit()
      mtNavigate()
      mtSave()
                                             

   Dependancies:
   *************
      Requires a table with the following structure to store the Memo Text.

      Memo_ID         Character   10  .0   None
      Memo_Sequence   Character    4  .0   None
      Memo_Text       Character  100  .0   None

      Requires that the rowset events for the table containing the link field to
      the memo text are set up to fire the control's corresponding methods.
      Along these lines:
      
.     function rowset_onAbandon
         form.MemoText1.mtAbandon()
         return

.     function rowset_onAppend
         form.MemoText1.mtAppend()
         return

.     function rowset_onDelete
         form.MemoText1.mtDelete()
         return

.     function rowset_onEdit
         form.MemoText1.mtEdit()
         return

.     function rowset_onNavigate(type, nRows)
         form.MemoText1.mtNavigate()
         return

.     function rowset_onSave
         form.MemoText1.mtSave()
         return


   Instantiation:
   ************** 
      In your form's Class code:

         set procedure to MemoText.cc additive

      Drop instance of the control onto the form. Ensure that the control's
      methods are called as detailed above.

      A Demo/Test form and tables ship with this control.
   

   Conditions of use:
   ******************
      This control is offered as "freeware" to dBASE and dB2K developers,
      but must be credited to the author.

      Since the use of this control is in data storage, it is the duty of the
      user to determine it's suitability for use. In no circumstance will the
      author accept any responsibility for any data loss, or any other problem
      which might arise from the use of this control.

      Feedback:

      If you find this control useful, the author would appreciate an e-mail
      to that effect.

      If you find any bugs or problems, the author would appreciate an e-mail
      with concise details. 
 
         
*/

class MemoText(parentObj) of EDITOR(parentObj) custom
   with (this)
      onChange = class::mtOnChange
      width = 200
      height = 240
      value = ""
      colorNormal = "black/Window"
      wrap = true
      modify = false
      parent.metric := 6 // Pixels
   endwith

   
   Protect datalink, onChange, onCLose
   

   function mtSetup(cTableName, oLinkedRowset, cLinkFieldName)
      // Attach passed params to control
      this.TableName = cTableName
      this.LinkedRowset = oLinkedRowset
      this.LinkFieldName = cLinkFieldName
      // Check for table and create if required
      if not file(this.TableName)
         if msgbox("The MemoText table " + this.Tablename + " cannot be found.";
          + chr(13) + "Do you wish a new empty table to be created ?","Missing Table",52) = 6
            private cNewTable
            cNewTable = "create table " + '"' + this.TableName + '"' + " (";
                 + "Memo_ID        char(10),";
                 + "Memo_Sequence  char(4),";
                 + "Memo_Text      char(100))"
            &cNewTable.
          else
             msgbox("You cannot continue error free without this table","Missing Table",16)
          endif
      endif
      // Setup Query
      sQueryString = "'select * from " + '"' + this.TableName + '"' + ;
         " where Memo_ID = :ID'"
      this.q = new QUERY()
      this.q.parent = this
      with (this.q)
         sql = &sQueryString.
         params["ID"] = ""
         active = true
      endwith
      // Populate control
      this.mtNavigate()
      return
    

   function mtOnChange()
      this.LinkedRowset.modified = true
      //
      // Deal with no pre-existing memo entry for this row
      this.q.rowset.first()
      if this.q.rowset.endofset
         local cNextID
         cNextID = this.Generate_ID()
         this.LinkedRowset.fields[this.LinkFieldName].value = cNextID
         this.LinkValue = cNextID
         this.q.rowset.beginappend()
      endif
      //
      return


   function mtOnClose()
      this.q.active = false
      this.q = null
      //
      // try to pack deleted rows
      try
         _app.databases[1].packTable(this.TableName)
      catch ( Exception e )
         // Do nothing - can't pack table without exclusive use.
         // Open project file prevents this in development environment.
         // Multiple controls will also prevent this until this event
         // fires for the last one, when the pack should succeed. 
      endTry
      //
      return

       
   function Generate_ID()
      return itoh(val(year(date()) ;
         + str(month(date()),2,0,"0") ;
         + str(day(date()),2,0,"0") ;
         + str(int((seconds() * 100)),7,0,"0")),8) ;
         + str((random()*100),2,0,"0")

            
   function mtNavigate()
      if not this.LinkedRowset.endofset
         if this.LinkedRowset.fields[this.LinkFieldName].value = null
            this.LinkValue = ""
         else
            this.LinkValue = this.LinkedRowset.fields[this.LinkFieldName].value
         endif
         this.q.params["ID"] = this.LinkValue
         this.q.requery()
         local sConcant
         sConcant = ""
         this.q.rowset.first()
         do while not this.q.rowset.endofset
            sConcant = sConcant + this.q.Rowset.Fields["Memo_Text"].value
            this.q.Rowset.next()
         enddo
         this.value = sConcant
      else
         this.value = ""
      endif
      return


   function mtEdit()
      this.modify = true
      return


   function mtSave()
      local sConcant
      sConcant = this.value
      //
      if len(this.value) = 0
         this.LinkedRowset.fields[this.LinkFieldName].value = ""
         this.LinkedRowset.save()
      endif
      //
      if not this.q.rowset.state = 3 // append
         this.q.rowset.first()
      endif
      //
      local nSeq, sMemoChunk
      nSeq = 1
      do while len(sConcant) > 0
         if this.q.rowset.endofset
            this.q.rowset.beginAppend()
         endif
         this.q.rowset.Fields["Memo_ID"].value = this.LinkValue
         this.q.rowset.Fields["Memo_Sequence"].value = str(nSeq,4,0,"0")
         sMemoChunk = substr(sConcant,1,100)
         this.q.rowset.Fields["Memo_Text"].value = sMemoChunk
         sConcant = substr(sConcant,101,len(sConcant))
         if len(sConcant) < 100
            sConcant = trim(sConcant)
         endif
         this.q.rowset.save()
         this.q.rowset.next()
         nSeq = nSeq + 1
      enddo
      // Clean up unused rows (shorter memo)
      do while not this.q.rowset.endofset
         this.q.rowset.delete() // this moves us to the next row
      enddo
      this.modify = false
      return


   function mtAbandon()
      this.modify = false
      this.q.rowset.abandon()
      this.mtNavigate()
      return


   function mtAppend()
      local cNextID
      cNextID = this.Generate_ID()
      this.LinkedRowset.fields[this.LinkFieldName].value = cNextID
      this.LinkValue = cNextID
      this.q.params["ID"] = cNextID
      this.q.requery()
      this.q.rowset.beginappend()
      this.value = ""
      this.modify = true
      return


   function mtDelete()
      this.q.rowset.first()
      do while not this.q.rowset.endofset
         this.q.rowset.delete() // this moves us to the next row
      enddo
      this.value = ""
      this.mtNavigate()
      return


endclass

